/*
     This file is part of libextractor.
     (C) 2004 Vidyut Samanta and Christian Grothoff

     libextractor is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     libextractor is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with libextractor; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.

     This code was based on AVInfo 1.0 alpha 11 
     (c) George Shuklin, gs]AT[shounen.ru, 2002-2004
     http://shounen.ru/soft/avinfo/

 **/

#include "extractor_util.h"
#include "pack.h"
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef MINGW
#include <sys/mman.h>
#include <netinet/in.h>
#endif

#define MPEG_DEBUG 0
#define HAVE_MAIN 0

static void addKeyword(struct EXTRACTOR_Keywords ** list,
		       char * keyword,
		       EXTRACTOR_KeywordType type) {
  EXTRACTOR_KeywordList * next;
  next = malloc(sizeof(EXTRACTOR_KeywordList));
  next->next = *list;    
  next->keyword = keyword;
  next->keywordType = type;
  *list = next;
}

#define MAX_READING_BYTES 65536
#define MPEG_ISO_11172_END_CODE 0x000001B9
#define MPEG_PACK_HEADER 0x000001BA
#define MPEG_SYSTEM_HEADER 0x000001BB
#define MPEG_PICTURE_SEQUENCE 0x00001B3
#define MAX_STREAMS 9

const unsigned short int mp3_bitrate_table[16][8]={
  /*                       MP3   MP2				*/
  /*v2l3  v2l2  v2l1  v2l? v1l3! v1l2  v1l1  */
  { 0,    0,    0,    0,    0,    0,    0,    0 },
  { 0,    8,   32,   32,    0,   32,   32,   32 },
  { 0,    16,  48,   64,    0,   40,   48,   64 },
  { 0,    24,  56,   96,    0,   48,   56,   96 },
  { 0,    32,  64,  128,    0,   56,   64,  128 },
  { 0,    64,  80,  160,    0,   64,   80,  160 },
  { 0,    80,  96,  192,    0,   80,   96,  192 },
  { 0,    56, 112,  224,    0,   96,  112,  224 },
  { 0,    64, 128,  256,    0,   112, 128,  256 },
  { 0,    128,160,  288,    0,   128, 160,  288 },
  { 0,    160,192,  320,    0,   160, 192,  320 },
  { 0,    112,224,  352,    0,   192, 224,  352 },
  { 0,    128,256,  384,    0,   224, 256,  384 },
  { 0,    256,320,  416,    0,   256, 320,  416 },
  { 0,    320,384,  448,    0,   320, 384,  448 },
  { 0,    0,   0,    0,     0,     0,   0,    0 }
};
unsigned short int mp3_freq_table[4][2]={
  {22050,44100},
  {24000,48000},
  {16000,32000},
  {0,    0}
};

const char fps_table[16][2]={ /*{0,0} means error or unknown value*/
  {0,0},{23,98},{24,0},{25,0},	
  {29,97},{30,0},{50,0},{59,94},
  {60,0},{0,0},{0,0},{0,0},
  {0,0},{0,0},{0,0},{0,0}
};

const char aspect_table[16][2]={ /*format: Y:X*/
  {0,0},{1,1},{3,4},{16,9},
  {0,0},{0,0},{0,0},{0,0},
  {0,0},{0,0},{0,0},{0,0},
  {0,0},{0,0},{0,0},{0,0}
};

typedef struct{
  int exist;
  int x;
  int y;
  int fps;
  int aspect;
  int bitrate;
} vids_t;

typedef struct{
  int exist;
  int bitrate;
  int freq;
  int ch;
} auds_t;

/* video/mpeg */
struct EXTRACTOR_Keywords * libextractor_mpeg_extract(char * filename,
						      char * xdata,
						      size_t xsize,
						      struct EXTRACTOR_Keywords * prev) {

#define GETC(a) do {if(pos<max_read) a=data[pos++]; else break;} while(0)
#define GETDW(a) do {if(pos<max_read-4) {a=data[pos++]<<24; a+=data[pos++]<<16;a+=data[pos++]<<8;a+=data[pos++];}else break;} while(0)
#define GETW(a) do {if(pos<max_read-2){ a=data[pos++]<<8;a+=data[pos++];} else break;} while(0)
#define SKIP(a) do{pos+=a;}while(0)
#define SKIPC() do{pos++;}while(0)
#define INSC(a) do {if(pos<max_read);a=(a<<8)+(data[pos++]);} while(0)

  unsigned char * data; /* unsigned is important! */
  size_t pos=0;
  size_t max_read;
  unsigned int sync=MPEG_ISO_11172_END_CODE;
  int c;
  int size;
  int size2=0;
  int temp=0;
  int common_bitrate=0;
  unsigned int vids_n=0;
  unsigned int auds_n=0;
  vids_t vids[MAX_STREAMS];
  auds_t auds[MAX_STREAMS];
  int mp3_mpeg_layer_ver=0;
  int mpeg2=0;

  data = (unsigned char*)xdata;
  for (c=0;c<MAX_STREAMS;c++){ 
    vids[c].exist=0; 
    auds[c].exist=0;
  }
  
  if (xsize > MAX_READING_BYTES)
    max_read = MAX_READING_BYTES;
  else
    max_read = xsize;

  if (max_read == 0)
    return prev;
  GETDW(sync);
  while(pos<max_read){
    if(sync==MPEG_ISO_11172_END_CODE){
#if MPEG_DEBUG
      printf ("got a EoF at 0x%x\n",pos-4);
#endif
      break;
    }
    if(sync==MPEG_PACK_HEADER){/*skip pack header*/
#if MPEG_DEBUG
      printf ("got a PACK_HEADER at 0x%x\n",pos-4);
#endif
      GETC(temp);
      if((temp&0xF0)==0x20) mpeg2=0;
      if((temp&0xC0)==0x40) mpeg2=1;
      if(mpeg2){
	SKIP(8);
	GETC(size);
	SKIP(size&3);
      } else SKIP(6);
      GETDW(sync);
      continue;
    }
    if(sync==MPEG_SYSTEM_HEADER){
#if MPEG_DEBUG
      printf ("got a SYSTEM_HEADER at 0x%x\n",pos-4);
#endif
      GETW(size);
      if (size < 0)
	return prev;
      SKIP(size);
      GETDW(sync);
      continue;
    }
    if ((sync&0xFFFFFC0)==0x0001C0) {/* audio or video PES header*/
#if MPEG_DEBUG
      printf ("got a a/v header at 0x%x, 0x%x\n",pos-4,sync);
#endif
      GETW(size); /*PES packet length*/
      if(!mpeg2){
	size2=0;
	do{
	  GETC(temp); size2++;
	  if (pos >= max_read)
	    return prev;
	} while (temp==0xFF);
	if ((temp&0xC0)==0x40) {
	  SKIP(1);
	  GETC(temp);
	  size2+=2;
	}
	if (temp&0x20)
	  SKIP(4);
	size2+=4;
	if(temp&0x10) 
	  SKIP(5);
	size2+=5;
      } else{/*if mpeg2*/
	SKIP(2);		
	GETC(size2); 
	if (size2 < 0)
	  return prev;
	SKIP(size2);
	size2+=3;
      }
      GETDW(temp); 
      if (sync&0x20){ /*video*/
	vids_n=sync&0xF;
#if MPEG_DEBUG
	printf("found video #%d, signature=0x%x\n",sync&0xF,temp);
#endif
	if (vids_n>MAX_STREAMS) {
	  return prev;
	}
	if( temp==MPEG_PICTURE_SEQUENCE){
	  if(pos>MAX_READING_BYTES-12) break;
	  vids[vids_n].exist=1;
	  vids[vids_n].exist=1;
	  vids[vids_n].x=(data[pos++]<<4);
	  vids[vids_n].y=((data[pos++]&0xF)<<8);
	  vids[vids_n].y+=(data[pos++]);
	  vids[vids_n].aspect=(data[pos]&0xF0)>>4;
	  vids[vids_n].fps=(data[pos++]&0x0F);
	  vids[vids_n].bitrate=((data[pos++]<<10)+(data[pos++]<<2)+(data[pos++]&0xC0))*400;
	  common_bitrate+=vids[vids_n].bitrate;
	}
      }else {
	auds_n=sync&0x1F;
#if MPEG_DEBUG
	printf("found audio #%d\n",sync&0x1F);
#endif
	if(auds_n>MAX_STREAMS) {
	  return prev;
	}
	if((temp&0xFFF00000)==0xFFF00000){
	  auds[auds_n].exist=1;
	  mp3_mpeg_layer_ver=(temp&0xE0000)>>17;
	  auds[auds_n].bitrate=mp3_bitrate_table[(temp&0xF000)>>12][mp3_mpeg_layer_ver]*1000;
	  auds[auds_n].freq=mp3_freq_table[(temp&0xC00)>>10][mp3_mpeg_layer_ver>>2];
	  auds[auds_n].ch=(temp&0xC0)==0xC0?1:2;
	  common_bitrate+=auds[auds_n].bitrate;
	}
      }
      if (size < size2 + 4)
	return prev; /* malformed */
      SKIP(size-size2-4);
      GETDW(sync);
      continue;
    }
    if ((sync&0xFFFFFF00)==0x100){ /* skip all other headers */      
      GETW(size);
      if (size <= 0)
	return prev; /* malformed */
      SKIP(size);
      GETDW(sync);
      continue;
    }
    INSC(sync);
  }

  size = xsize;

  temp=(int)((double)size*8/(double)(common_bitrate?common_bitrate:0xFFFFFFFF));
  for(c=0;c<MAX_STREAMS;c++){
    char * format;
    if (auds[c].exist) {
      format = malloc(256);
      snprintf(format,
	       256,
	       "audio: %u hz, %u kbps, %dm%d, %s",
	       auds[c].freq,
	       auds[c].bitrate / 1000,
	       temp / 60, temp % 60,
	       auds[c].ch > 1 ? "stereo" : "mono");
      addKeyword(&prev,
		 format,
		 EXTRACTOR_FORMAT);
    }
    if(vids[c].exist) {      
      addKeyword(&prev,
		 strdup("video/mpeg"),
		 EXTRACTOR_MIMETYPE);
      format = malloc(256);
      snprintf(format,
	       256,
	       "video: %d.%d fps at %d kbps, %dm%d",	     
	       fps_table[vids[c].fps][0],
	       fps_table[vids[c].fps][1],
	       vids[c].bitrate / 1024,
	       temp / 60, temp % 60);
      addKeyword(&prev,
		 format,
		 EXTRACTOR_FORMAT);
      format = malloc(256);
      snprintf(format,
	       256,
	       "%ux%u",
	       vids[c].x,
	       vids[c].y);
      addKeyword(&prev,
		 format,
		 EXTRACTOR_SIZE);
    /*
      SetIdxNumericVar("v%d.aspectY",c+1,aspect_table[vids[c].aspect][0]);
      SetIdxNumericVar("v%d.aspectX",c+1,aspect_table[vids[c].aspect][1]);
      SetIdxNumericVar("v%d.cc",c+1,mpeg2?0x3247504D:0x3147504D);
      */
    }
  }
  return prev;
}




#if HAVE_MAIN
int main (int argc, char **argv) {
  int file;
  char * buffer;
  struct stat fstatbuf;
  size_t size;

  if (argc != 2) {
    fprintf(stderr, 
	    "Call with filename as argument\n");
    return -1;
  }  
  file = OPEN(argv[1],O_RDONLY);
  if (-1 == file) 
    return -1;
  if (-1 == fstat(file, &fstatbuf)) {
    close(file);
    return -1;
  }
  size = fstatbuf.st_size;
  buffer = mmap(NULL, size, PROT_READ, MAP_SHARED, file, 0);  
  close(file);

  EXTRACTOR_printKeywords(stdout,
			  libextractor_mpeg_extract(argv[1], 
						    buffer,
						    size,
						    NULL));
  munmap(buffer,size);
  return 0;
}
#endif
